"""
   Addr2Line2Locations-for-gps (version 1, revision 2, 2010-01-17)
   ---------------------------------------------------------------------------

   Provide invok_addr2line (executable_name = None, exception_log = None),
   which will invok addr2line on each addresse parsed from
   exception_log, using executable_name for the "--exe" parameter.

   executable_name is the name of an executable, with extension if
   applicable and full path if required.

   exception_log is the output of an exception log as returned by an
   Ada application compiled with GNAT, preferably with exception traceback
   enabled at build time (without traceback, only the exception origin
   location will be show, if possible, because it may not be present).

   If one or both of executable_name and exception_log is None, then
   opens a dialog using request_addr2line_parameters_input, which will
   direct the user to fill the requested informations (this try to guess
   the possible executable names from the project).

   When executable_name and exception_log are available, then, run addr2line
   as explained before, adding entries in the Locations pane of GPS. Entries
   are added under a category name which use the base name of the executable
   with extension.

   Whenever an error occured, a console window is opened, which report
   errors from addr2line. If all invocation failed (there will be tipically
   multiple invocations), then the user is warned with a dialog box, telling
   him something is probably really wrond (missing debugging informations
   as an example).

   If the dialog is involved in the process (if either one or both of the
   two parameters is None), then the user may cancel the process in this
   dialog. Otherwise, the process of invoking addr2line iteratively is
   started immediately.

   As entries in the Locations pane are organized by source file name,
   the description text for each entry shows an index which is the index
   of the addresse in the exception traceback.

   The main method, invok_addr2line, may be invoked as-is from a menu
   entry, without parameters or invoked from any other module, with or
   without parameters.

   For GPS, this will add an entry in the Tools menu. This is added to
   the Tools menu, because unlike all entries in the Debug menu, this has
   nothing to deal with GDB, and the entry is enabled all the time,
   even when no executable is currently beeing debugged. The sub
   menu Build/Run, is probably not a good idea either. So the choice was
   made to add it to Tools.

   Last version and oneline documentations :
   (en) http://www.les-ziboux.rasama.org/addr2line2locations-tool-ada-gps-en.html 
   (fr) http://www.les-ziboux.rasama.org/addr2line2locations-tool-ada-gps.html

   ---------------------------------------------------------------------------
   January 2010 - Metz (France) - Contact, suggestions and comments :
   yannick_duchene@yahoo.fr or else yannick-duchene@rasama.org
   This application may be redistributed, modified or not, for commercial
   purpose or not, at your option, under the terms of the Two-Clause
   BSD License (Simplified BSD License). This copyright notice and contact
   informations must be preserved.
   ---------------------------------------------------------------------------
"""


import gtk
import os
import os.path
import subprocess
import sys
import re
import GPS

# ============================================================================
# TOC
# ----------------------------------------------------------------------------
# Text table
# message_box (icon_stock_id, title, message)
# error_message_box (title, message)
# warning_message_box (title, message)
# information_message_box (title, message)
# get_addresses_or_kind_from_exception_log (exception_log)
# platform_executable_extension ()
# project_executables_names ()
# category_name (executable_name)
# get_active_console ()
# get_exception_message_from_active_console ()
# request_addr2line_parameters_input (...) *helper for the main method*
# is_decimal_digit (c)
# invok_addr2line_with_parameters (executable_name, address, index = None)
# get_exception_origin_location_from_exception_log (exception_log)
# invok_addr2line (executable_name = None, exception_log = None) *main method*
# on_click_invok_addr2line (ui_control)
# on_UI_ready (hook_name)
# integration into GPS
# Output samples from Addr2Line (for output interpretation reference)
# Samples exception messages
# Change log
# ----------------------------------------------------------------------------

# ============================================================================
# Text table

# You may translate these, whenever you wish a full localization of this
# application. No pre-localized version of this application is provided,
# for the sole reason near to all of peoples concerned by this kind
# of application, will read english any way.

locations_category_name_pattern = (
   "Exception traceback for $executable_name")

exception_origin_location_pattern = (
   "Origin of exception $exception_name")

application_console_name = (
   "Addr2Line2Locations")

application_title = (
   "Addr To Line To Locations")

all_addr2line_invocations_failed_message = (
   "All invocations of addr2line have failed. Check either that " +
   "the executable to inspect was found and its name or its path was not " +
   "misspelled, the executable was compiled and built with sufficient " +
   "debugging informations which was not stripped at link time, and " +
   "last but not least, addr2line is present on your platform and " +
   "it's in your ${PATH}.")

location_description_pattern_with_index = (
   "The #$index from inner most at address 16#$address#")

location_description_pattern_without_index = (
   "At address 16#$address#")

exception_trace_help_text = (
   "If not properly guessed, paste here the end of the output of the " +
   "application which returned you an exception log. This is the part " +
   "of the ouput which starts with a \"Execution terminated by unhandled " +
   "exception\" or with line telling you \"raised <EXCEPTION_NAME>\"" +
   "and which preferably contains a list hexadecimal numbers using the " +
   "C notation (i.e. like 0xHHHHHHHH).")

executable_name_help_text = (
   "Edit or select here, the name of the application executable wich " +
   "returned you an exception log. The name provided here must be well " +
   "suited as an \"-e\" or \"--exe\" parameter for addr2line. Directory " +
   "specification is not required if the executable is in the project " +
   "current directory.")

exception_log_must_not_be_empty = (
   "The exception log text you have to paste must not be empty. " +
   "Please, edit again or else cancel the dialog.")

executable_name_must_not_be_empty = (
   "The executable name you have to paste or edit, must not be empty. " +
   "Please, edit again or else cancel the dialog.")

executable_name_must_be_single_line = (
   "The executable name, is a name, and thus, it cannot span through " +
   "multiple lines. Please, edit again or else cancel the dialog.")
   
no_addresses_found_in_exception_log = (
   "No traceback addresses could be found in the provided exception log " +
   "text. Nothing looking like an exception message coild been recognize " +
   "either. Please, edit again, or else cancel the dialog.")
   
exception_log_message_title = (
   "About the exception log text")

executable_name_message_title = (
   "About the executable name")

internal_error_message = (
   "An internal error occured in \"" +
   application_title +
   "\"")

unknown_exception_origin_message = (
   "Note: the location of the statement which raised " +
   "the exception is unknow")

no_location_informations_at_all_message = (
   "No location informations about the exception could be retrieved " +
   "at all ; nor its origin location nor any traceback locations.")

# ============================================================================

def message_box (icon_stock_id, title, message):
   """
      + message_box (icon_stock_id, title, message):
      - Opens a dialog box as a message box.
      - <icon_stock_id> will be used for an image displayed in the dialog's
        action_area, at the left of the Close button.
      - <title> will be the dialog title.
      - <message> will be the text buffer content of a read-only text view.
      - See also : error_message_box and information_message_box
   """

   # -------------------------------------------------------------------------
   # TOC
   # -------------------------------------------------------------------------
   # Dialog box
   # Message box icon
   # Button
   # Message view
   # Run
   # -------------------------------------------------------------------------
      
   # -------------------------------------------------------------------------
   # Dialog box

   dialog = gtk.Dialog (title)

   # -------------------------------------------------------------------------
   # Message box icon

   message_icon = gtk.Image ()
   message_icon.set_from_stock (
      stock_id = icon_stock_id,
      size = gtk.ICON_SIZE_DND)
   message_icon.show ()
   dialog.action_area.add (message_icon)

   # -------------------------------------------------------------------------
   # Button

   close_button = dialog.add_button (gtk.STOCK_CLOSE, 1)

   # -------------------------------------------------------------------------
   # Message view

   message_view = gtk.TextView ()
   message_view.set_border_width (10)
   message_view.set_editable (False)
   message_view.set_sensitive (False)
   message_view.set_wrap_mode (gtk.WRAP_WORD)
   message_view.set_size_request (width = 300, height = 93)
   message_view.get_buffer ().set_text (message)
   message_view.show ()
   dialog.vbox.add (message_view)

   # -------------------------------------------------------------------------
   # Run

   dialog.run ()
   dialog.destroy ()

# ============================================================================

def error_message_box (title, message):
   """
      + error_message_box (title, message):
      - Invok <message_box> (see message_box) with gtk.STOCK_DIALOG_ERROR as
        the icon_stock_id, and title and message as the title and message
        arguments.
      - See also : information_message_box
   """

   message_box (
      icon_stock_id = gtk.STOCK_DIALOG_ERROR,
      title         = title,
      message       = message)

# ============================================================================

def warning_message_box (title, message):
   """
      + information_message_box (title, message):
      - Invok <message_box> (see message_box) with gtk.STOCK_DIALOG_WARNING as
        the icon_stock_id, and title and message as the title and message
        arguments.
      - See also : error_message_box
   """

   message_box (
      icon_stock_id = gtk.STOCK_DIALOG_WARNING,
      title         = title,
      message       = message)

# ============================================================================

def information_message_box (title, message):
   """
      + information_message_box (title, message):
      - Invok <message_box> (see message_box) with gtk.STOCK_DIALOG_INFO as
        the icon_stock_id, and title and message as the title and message
        arguments.
      - See also : error_message_box
   """

   message_box (
      icon_stock_id = gtk.STOCK_DIALOG_INFO,
      title         = title,
      message       = message)

# ============================================================================

an_exception_log_without_trace_back = 1
# This could be any thing, this is just a value
# to compare to.

def get_addresses_or_kind_from_exception_log (exception_log):
   """
      + get_addresses_or_kind_from_exception_log (exception_log):
      - Returns either a list, None, or an integer
      - Returns a list of addresses from <exception_log>.
      - Attention: if the list would be empty, returns None instead.
      - Returns <an_exception_log_without_trace_back> if no traceback were
        found but the exception log still stands for something which
        may be used : an exception message without trace back.
      - <exception_log> is splitted into words, and then words
        matching the form "0x??" where "??" is
        at least two characters long, are added to the result.
      - <exception_log> is either the output of an application
        which terminates which an exception log, or the end
        of its output, which will be parsed as possible.
      - If the string "Call stack traceback locations:" is found,
        then only the substring starting after this
        marker is taken into account.
      - This is used by <request_addr2line_parameters_input> and
        by <invok_addr2line>.
      - Note: the items in the list are not cheched to be really
        composed of hexadecimal digits, only the "0x" prefix is checked for
        and the lenth which ensure if will contains at least 2 digits.
   """

   # -------------------------------------------------------------------------

   # Split after one of exception log tipycal markers if found.
   # This is useful to prevent parsing of adresses from what may
   # not be addresses. If the marker is not found, then the text
   # is supposed to be a manual copy/paste of the exception log
   # text, and for this reason, it is then parse entirely.

   marker = "Call stack traceback locations:"
   # See the comments in the section named "Samples exception messages" at
   # the end of this file to understand this.

   trace_marker_position = exception_log.find (marker)

   if trace_marker_position > -1:
      trace_marker_position = trace_marker_position + len (marker)
      # Do not keep the marker, just what starts right after it.
      final_text = (
         exception_log [trace_marker_position + 1:len (exception_log)])
   else:
      final_text = exception_log

   # -------------------------------------------------------------------------

   # See the comments in the section named "Samples exception messages" at
   # the end of this file to understand this.

   words = final_text.split (" ")
   addresses = []

   for word in words:
      if len (word) >= len ("0xHH"):
         if (word[0] == "0") and (word[1] == "x"):
            addresses.append (word)
            
   if len (addresses) > 0:
      return addresses
   elif exception_log.find ("raised ") > -1:
      return an_exception_log_without_trace_back
   else:
      return None
      
# ============================================================================

def platform_executable_extension ():
   """
      + platform_executable_extension ():
      - Returns a string which is stands for the suffix to
        be added to executable names on the platform.
      - If the platform has no standard extension for executable,
        then returns an empty string.
      - If the platform has a standard extension for executable,
        then return the extension, including the dot.
      - The result of this function is to be appended directly to
        any executable base file name without extension.
      - Used by <project_executables_names> to get real
        executable file name from file names without extensions.
   """

   # We use a trick : get the extension of the Python executable.
   # Otherwise, there is not such dedicated function in Python
   # standard modules (thus the trick).
   
   python_executable = os.path.basename (sys.executable)
   
   file_name_parts = os.path.splitext (python_executable)
   
   if file_name_parts[1] == "":
      return ""
   elif file_name_parts[1] == ".":
      return ""
   else:
      return file_name_parts[1]
   
# ============================================================================

def project_executables_names ():
   """
      + project_executables_names ():
      - Returns the list of executables names produced by a project.
      - This first read main files from the project, then
        for each main file, apply the executable renaiming defined
        by the Builder package, if there are some, otherise, use
        the main file basename without extension as the executable
        name.
      - Returns the result as an array of strings with the platform
        executable extension appended if applicable.
      - It is not guaranted the result may not be an empty list.
      - Used by <request_addr2line_parameters_input> to ease user
        input.
   """

   result = []

   # -------------------------------------------------------------------------

   project = GPS.Project.root ()

   # -------------------------------------------------------------------------

   # Get the list of mains file defined in the project. Main files
   # are the one which will produce executables.

   mains = project.get_attribute_as_list (
      package   = "",
      attribute = "Main",
      index     = "")

   # -------------------------------------------------------------------------

   # Executables may be renamed, as allowed by the GPR project syntax.
   # So now we have to handle all defined renaming. If no renaming if
   # found, then this means the executable will be named after its
   # source file name.

   for main in mains:
      executable_renaming = project.get_attribute_as_string (
         package   = "Builder",
         attribute = "Executable",
         index     = main)

      if executable_renaming != "":
         executable_name = executable_renaming
      else:
         executable_name = os.path.splitext (os.path.basename (main))[0]

      result.append (executable_name + platform_executable_extension ())

   # -------------------------------------------------------------------------

   return result

# ============================================================================

def category_name (executable_name):
   """
      + category_name (executable_name):
      - Returns the category name which will be used in the
        Locations pane.
      - This is maid a function, because must be shared between
        multiple parts of the application : the category name is
        not just needed by invok_addr2line_with_parameters, but also
        by some others parts which may add other informational
        entries in the output, as available.
      - Basically, the category name is based on the executable
        name.
   """

   executable_base_name = os.path.basename (executable_name)

   global locations_category_name_pattern

   return locations_category_name_pattern.replace (
      "$executable_name",
      executable_base_name)
   
# ============================================================================

def get_active_console ():
   """
      + get_active_console ():
      - Returns the active console if there is one, or
        else return None
      - This is a utility methods which provide something missing
        in GPS (well, by the way, this is feasible).
   """

   try:
      active_window = GPS.MDI.current ()
      # The specification of this methods states that if
      # there is no active window, then it raise an exception.
      # This is why this block is wrapped in a try statement.

      child = active_window.get_child ()
      # The given window cannot be tested as beeing a GPS.Console
      # or not. If it is the console, the console is contained the
      # window, as its single child.

      if isinstance (child, GPS.Console):
         # The active window does not contain a console, so
         # there is no active console.
         return child
      else:
         return None
      
   except:
      return None
   
# ============================================================================

def get_exception_message_from_active_console ():
   """
      + get_exception_message_from_active_console ():
      - Try to gets text from an active console, if no text
        could be found, returns an empty string, else,
        find a marker which is tipycal of an exception
        message and return the part of text starting at
        there, or else, return the whole text.
      - Used as a fallback initialization in
        <request_addr2line_parameters_input>
   """

   # -------------------------------------------------------------------------

   # Get a the active console or else return an empty string

   console = get_active_console ()

   if console == None:
      return ""

   # -------------------------------------------------------------------------

   text = console.get_text ()

   # -------------------------------------------------------------------------

   # See the section named "Samples exception messages" at the end
   # of this file, for samples exception messages and comments on
   # the strategy used to handle these. Reading these comments
   # will help you understand the following testes.

   start = text.find ("\nExecution terminated by unhandled exception")

   if start > -1:
      return (text[start:]).strip ()

   start = text.find ("\nraised ") # Notice the leading space

   if start > -1:
      return (text[start:]).strip ()

   return text.strip ()

# ============================================================================

def request_addr2line_parameters_input (
   initial_executable_name = None,
   initial_exception_log = None):
   """
      + request_addr2line_parameters_input (
           initial_executable_name = None,
           initial_exception_log = None):
      - Opens a dialog box which direct the user to input parameters
        to compute argument to invok addr2line.
      - Returns None if the user canceled the dialog
      - Returns a tuple of three items.
      - The first item of the result is a none-blank single line string
        which is the executable name.
      - The second item of the result, is a none-empty list of addresse
        as returned by <get_addresses_or_kind_from_exception_log>. See
        <get_addresses_or_kind_from_exception_log> for more details about the
        list.
      - The third item, is the exception log text which was the source of the
        second item (the addresses list). The exception log may be useful to
        get apart, to parse further more informations than the list of
        addresses.
      - The client of this function should use the addresses list to
        generate as much invocations to addr2line and the second item should
        be used as the "-e" or "--exe" parameter of addr2line.
      - If <initial_executable_name> is not None, then it is used as the
        initial content of the executable name input.
      - If <initial_exception_log> is not None, then it is used as the
        initial content of the exception log text input.
   """

   global application_title
   global exception_trace_help_text
   global executable_name_help_text
   global exception_log_must_not_be_empty
   global executable_name_must_not_be_empty
   global executable_name_must_be_single_line
   global no_addresses_found_in_exception_log
   global exception_log_message_title
   global executable_name_message_title
   global internal_error_message

   # -------------------------------------------------------------------------
   # TOC
   # -------------------------------------------------------------------------
   # Dialog box
   # Buttons
   # Exception log text purpose
   # Exception log text
   # Separator
   # Executable name purpose
   # Executable name text
   # Run the dialog
   # -------------------------------------------------------------------------

   # -------------------------------------------------------------------------
   # Dialog box

   dialog = gtk.Dialog (application_title)

   # -------------------------------------------------------------------------
   # Buttons

   # Create validation and cancel buttons in the dialog box action area.
   # No need to do a show() on buttons, as it is automatically done.

   ok_button = dialog.add_button (gtk.STOCK_OK, 1)
   cancel_button = dialog.add_button (gtk.STOCK_CANCEL, 0)

   # -------------------------------------------------------------------------
   # Exception log text purpose

   # Display a purpose help text as a read only text view.
   # It is added priorly to the input text view to be on top of it
   # (help comes before the control). The help text view is created with
   # a default text buffer (no buffer passed when the text view is created).

   exception_trace_help = gtk.TextView ()
   exception_trace_help.set_border_width (10)
   exception_trace_help.set_editable (False)
   exception_trace_help.set_sensitive (False)
   exception_trace_help.set_wrap_mode (gtk.WRAP_WORD)
   exception_trace_help.get_buffer().set_text (exception_trace_help_text)
   exception_trace_help.show ()
   
   dialog.vbox.add (exception_trace_help)

   # -------------------------------------------------------------------------
   # Exception log text

   # A text view, with a default text buffer (a default text buffer is
   # created because none is provided when the text view is created).

   exception_log_view = gtk.TextView ()
   exception_log_view.set_border_width (10)
   exception_log_view.set_editable (True)
   exception_log_view.set_sensitive (True)
   exception_log_view.set_wrap_mode (gtk.WRAP_WORD)
   exception_log_view.set_size_request (width = 300, height = 185)
   exception_log_view.show ()

   # Set initial content of exception log text : first try the exception_log
   # parameter, if it is None, try to get text from the active console, which
   # will be any way, limited to its last lines (because it may possibly be
   # big). If not text could be get from an active console,
   # get_text_from_active_console returns an empty string.

   exception_log_buffer = exception_log_view.get_buffer ()

   if initial_exception_log != None:
      exception_log_buffer.set_text (initial_exception_log)
   else:
      exception_log_buffer.set_text (
         get_exception_message_from_active_console ())

   dialog.vbox.add (exception_log_view)

   # -------------------------------------------------------------------------
   # Separator

   # Add a separator before the next section which will follow

   trace_to_executable_name_separator = gtk.HSeparator ()
   trace_to_executable_name_separator.show ()

   dialog.vbox.add (trace_to_executable_name_separator)

   # -------------------------------------------------------------------------
   # Executable name purpose
   
   # Executable name help

   executable_name_help = gtk.TextView ()
   executable_name_help.set_border_width (10)
   executable_name_help.set_editable (False)
   executable_name_help.set_sensitive (False)
   executable_name_help.set_wrap_mode (gtk.WRAP_WORD)
   executable_name_help.get_buffer().set_text (executable_name_help_text)
   executable_name_help.show ()

   dialog.vbox.add (executable_name_help)

   # -------------------------------------------------------------------------
   # Executable name text

   # Create a combo box whose item field is editable

   known_executables = []

   if initial_executable_name != None:
      known_executables.append (initial_executable_name)

   executables_from_project = project_executables_names ()
   for executable_name in executables_from_project:
      known_executables.append (executable_name)

   executable_name_text_view = gtk.Combo ()
   executable_name_text_view.set_border_width (10)
   executable_name_text_view.set_sensitive (True)
   executable_name_text_view.set_use_arrows (True)
   executable_name_text_view.set_use_arrows_always (True)
   executable_name_text_view.set_case_sensitive (True)
   executable_name_text_view.set_popdown_strings (known_executables)

   executable_name_text_view.set_value_in_list (
      val = False, # The user if free to edit a custom value
      ok_if_empty = False) # No empty value allowed
   
   executable_name_text_view.show ()

   dialog.vbox.add (executable_name_text_view)

   # -------------------------------------------------------------------------
   # Run the dialog

   # Getting texts may fail, this is why the dialog.run () statement is
   # inside a loop : we may have to re-run the dialog.

   while True:

      status = dialog.run ()

      if status == 1:

         # Getting text is wrapped in secure block, because it will
         # fail whenever the text is edited while we attempt to retrieve it.
         # If it fails, we simply re-run the dialog. We will also have
         # to re-run the dialog if the user made a wrong input (if so, then
         # he/she will be priorly notified about it).

         try:
            text_buffer = exception_log_view.get_buffer ()
            start_of_text = text_buffer.get_start_iter ()
            end_of_text = text_buffer.get_end_iter ()

            exception_log = text_buffer.get_text (
               start = start_of_text,
               end   = end_of_text,
               include_hidden_chars = False)

            executable_name = executable_name_text_view.entry.get_text ()

            exception_log = exception_log.strip ()
            executable_name = executable_name.strip ()

            # Test for validity condition missmatch, else return the result

            if exception_log == "":
               error_message_box (
                  title   = exception_log_message_title,
                  message = exception_log_must_not_be_empty)
            elif executable_name == "":
               error_message_box (
                  title   = executable_name_message_title,
                  message = executable_name_must_not_be_empty)
            elif executable_name.find ("\n") != -1:
               error_message_box (
                  title   = executable_name_message_title,
                  message = executable_name_must_be_single_line)
            else:
               result = get_addresses_or_kind_from_exception_log (
                  exception_log)

               if result == None:
                  error_message_box (
                     title   = exception_log_message_title,
                     message = no_addresses_found_in_exception_log)
               elif result == an_exception_log_without_trace_back:
                  dialog.destroy ()
                  return (executable_name, result, exception_log)
               elif type (result).__name__ == "list":
                  dialog.destroy ()
                  addresses = result
                  return (executable_name, addresses, exception_log)
               else:
                  # Should never occurs. Handled here for
                  # completness and safety.
                  error_message_box (
                     title   = application_title,
                     message = internal_error_message)
                  # Inform the user (so that the user can know
                  # and prefer to cancel the dialog right after),
                  # and re-run the dialog.
                  None

         except:
            None
            # Do nothing and simply re-run the dialog again

      else:
         # The user canceled the dialog
         dialog.destroy ()
         return None

# ============================================================================

def is_decimal_digit (c):
   """
      + is_decimal_digit (c):
      - Returns true, if <c>, which is a character,
        is a decimal digit, that is, it's in the range '0' .. '9'
   """

   decimal_digits = "0123456789"
   return decimal_digits.find (c) > -1

# ============================================================================

def invok_addr2line_with_parameters (executable_name, address, index = None):
   """
      + invok_addr2line_with_parameters (
           executable_name,
           address,
           index = None):
      - Invok addr2line with <executable_name> as the "--exe" parameter
        and <address> as the (anonymous) address parameter. The
        <executable_name> is enclosed with quotes, to be compatible with
        Windows plateforms, where an executable name may contains spaces.
      - The output of addr2line is parsed and if it stands for a successfull
        execution of addr2line, then a location is added in the GPS Locations
        pane, using the executable base name as the category name and the
        function returns True.
      - If addr2line did not successfully returned anything, nothing is added
        in the Locations pane, and the function returns False (reciprocally,
        think that when it returns False, nothing were added in the Locations
        pane).
      - The client of this function should cumulate the returned boolean
        value of all of the invocation, if it is to invok it multiple times,
        and warn the user if all invocation returned False. This may mean
        that the executable could not be found at all, or that the executable
        does not contains the required debugging informations to map from
        addresse to source location. In the worst case, this may mean
        addr2line is not present or not in the user's ${PATH}.
      - To go further, if addr2line failed, a GPS console is opened, which
        displays the message addr2line wrote on its STDERR stream.
      - If <index> is not None, then show this index in the text of the
        entry which is added to the Locations pane. As explained in the
        documentation of this module, this is used because the GPS
        Locations pane organize entries by source file name, thus, we
        need a way to show what was the real order of the exception
        trace points.
   """

   global location_description_pattern_with_index
   global location_description_pattern_without_index
   global application_console_name

   # -------------------------------------------------------------------------
   # TOC
   # -------------------------------------------------------------------------
   # Addr2Line process
   # Addr2Line's output interpretation
   # -------------------------------------------------------------------------

   # -------------------------------------------------------------------------
   # Addr2Line process

   command_line = "addr2line --exe=\"" + executable_name + "\" " + address
   # executable_name is surrounded by quotes, to ensure it will work
   # on all platforms. Ex. on Windows, a path may contains space.
   
   if subprocess.STARTUPINFO and subprocess.STARTF_USESHOWWINDOW:
      # On Windows platforms, avoid to open a console when the process
      # is launched. The test in the if-condition just check if we are
      # on Windows like platform.
      startupinfo = subprocess.STARTUPINFO()
      startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
   else:
      startupinfo = None

   addr2line_process = subprocess.Popen (
      args      = command_line,
      stdout    = subprocess.PIPE,
      stderr    = subprocess.PIPE,
      shell     = False,
      universal_newlines = True,
      startupinfo = startupinfo)

   # -------------------------------------------------------------------------
   # Addr2Line's output interpretation

   # See output samples from Addr2Line at the end of this file.
   # This samples are used as reference for the output interpretation
   # and are given here for completness. Interpretation strategy is also
   # explained there (in the comment at the end of this file).

   (stdout, stderr) = addr2line_process.communicate ()

   if addr2line_process.returncode == 0:

      stdout = stdout.strip ()
      # Strip, to remove heading and leading new line characters

      words = stdout.split (" ")

      # About these testes, see the comment at the end of the file

      if len (words) < 3:
         return False
      if len (words[0]) < 8:
         return False
      if (words[1] != "at") and (words[1] != "in"):
         return False
   
      last_word = words [len (words) - 1]
      last_character = last_word [len (last_word) - 1]
      if not is_decimal_digit (last_character):
         return False
      if words[2].find ("?") != -1:
         return False
      last_word = last_word[::-1]
      found_none_zero = False
      for digit in last_word:
         if not is_decimal_digit (digit):
            return False
         if digit != '0':
            found_none_zero = True
            break
      if not found_none_zero:
            return False

      if index != None:
         address_text = words[0]
         index_text = str (index)
         entry_text = location_description_pattern_with_index
         entry_text = entry_text.replace ("$index", index_text)
         entry_text = entry_text.replace ("$address", address_text)
      else:
         address_text = words[0]
         entry_text = location_description_pattern_without_index
         entry_text = entry_text.replace ("$address", address_text)

      start_of_path = len (words[0]) + 1 + len (words[1]) + 1
      path_with_line_number = stdout[start_of_path:len (stdout)]

      text_for_locations_parse = (
         path_with_line_number +
         ":1: " + # Notice the space, required by Locations.parse()
         entry_text)

      GPS.Locations.parse (
         output   = text_for_locations_parse,
         category = category_name (executable_name))

      return True

   else:
      console = GPS.Console (application_console_name)
      console.pywidget ().grab_focus ()
      console.clear ()
      # No use to keep previous error message. Further more,
      # these may be error messages from a previous <invok_addr2line>
      # invocation, and would then be confusing for the user.
      console.write (stderr)
      return False

# ============================================================================

def get_exception_origin_location_from_exception_log (exception_log):
   """
      + get_exception_origin_location_from_exception_log (exception_log):
      - An exception message from a application which terminated
        with an unhandled exception, may contains a call stack
        trace back or not. Any way, when this trace back is
        present, it does not contain the location of the statement
        which raised the exception (indeed, the trace back provides
        addresses in the call stack). This information, is located
        elsewhere in the exception message, and may be present
        even when no back trace is provided. So this information
        is independent of the trace back.
      - Returns the location of the statement which raised the exception,
        in a format which is parsable by GPS.Locations.parse. It
        contains the name of the exception, a file name and line
        number. It seems the file name is always relative to
        a project and never includes directory information. So
        the Locations pane will be able to open the location only
        if the file is part of the project (this may be via
        a dependency project).
      - Returns None is no location was found. This occurs
        whenever an exception is raised with a custome message, because
        then, the custom message replace the file name and line
        information.
   """

   global exception_origin_location_pattern

   # Read the section named "Samples exception messages" at the end
   # of this file, to understand the implementation of this
   # method.

   # Note : GNAT does not allow space is source file names,
   # so we do not handle this case (remmember there is no
   # directory, just filename, so we don't have do handle
   # space even on platforms which make use of space in file names).

   # First attempt with the "raised " kind of message

   marker = "raised "
   # The source location, if found, will be after the
   # a separator on the same, up to the end of line.
   # The exception name will be after the marker, up to
   # the separator. See below.

   lines = re.findall (
      pattern = marker + ".+",
      # The DOTALL flag is not specified, so the '.' does not match
      # new line. Match will end at end of line.
      string  = exception_log,
      flags   = re.MULTILINE)

   if len (lines) > 0:
      # We have something of the kind
      # raised EXCEPTION_NAME : a4g-vcheck.adb:885
      # or
      # raised EXCEPTION_NAME : Custom message
      line = lines[0]
      separator = " : "
      separator_position = line.find (separator)
      if separator_position == -1:
         return None
      start = separator_position + len (separator)
      # What we want to keep, is located after the separator
      limit = len (line)
      if not is_decimal_digit (line [limit - 1]):
         return None
      location = line[start:limit]
      location = location + ":1: "
      # Notice the space, required by GPS.Locations.parse
      # Now, get the exception name
      start = len (marker)
      limit = separator_position
      # What we want to keep now, if before the separator, so
      # it ends at the separator
      exception_name = line[start:limit]
      location = (
         location +
         exception_origin_location_pattern.replace (
            "$exception_name",
            exception_name))
      return location

   # Second attempt with the "Message: " kind of message

   marker = "Message: "
   # The source location, if found, will be after the
   # marker, up to the end of line.
   # The exception name will start after another marker
   # and up to the end of line. See below.

   lines = re.findall (
      pattern = marker + ".+",
      string  = exception_log,
      flags   = re.MULTILINE)

   if len (lines) > 0:
      # We have something of the kind
      # (...)
      # Exception name: ASIS.EXCEPTIONS.ASIS_INAPPROPRIATE_ELEMENT
      # Message: a4g-vcheck.adb:885
      # or
      # (...)
      # Exception name: ASIS.EXCEPTIONS.ASIS_INAPPROPRIATE_ELEMENT
      # Message: Custom message

      line = lines[0]
      start = len (marker)
      limit = len (line)
      if not is_decimal_digit (line [limit - 1]):
         return None
      location = line[start:limit]
      location = location + ":1: "
      # Notice the space, required by GPS.Locations.parse

      # Now, get the exception name, which is
      # marked with another marker.

      marker = "Exception name: "

      lines = re.findall (
         pattern = marker + ".+",
         string  = exception_log,
         flags   = re.MULTILINE)

      if len (lines) == 0:
         return None

      line = lines[0]
      start = len (marker)
      limit = len (line)
      exception_name = line[start:limit]
      location = (
         location +
         exception_origin_location_pattern.replace (
            "$exception_name",
            exception_name))
      return location
      
   return None

# ============================================================================

exception_origin_location_added = False

def add_exception_origin_location (executable_name, exception_log):
   """
      + add_exception_origin_location (executable_name, exception_log):
      - Add an enty in the Locations pane, for the location of
        the statement which raised the exception.
      - If this location is unknown, then inform the user. It is
        mostly required to inform the user, because if there is
        no exception origin location and in the mean time, no
        trace back information, then the user will get no feed-back
        at all from the application, which is a very bad thing for
        a user/application interface.
      - If an exception location could be added, set
        <exception_origin_location_added> to True. Reset it
        to False otherwise (this is a procedure, not a function,
        this is why it does not return a value and use a global
        variable instead).
   """

   global application_title
   global unknown_exception_origin_message
   global exception_origin_location_added

   exception_origin_location = (
      get_exception_origin_location_from_exception_log (
         exception_log))

   if exception_origin_location != None:
      GPS.Locations.parse (
         output   = exception_origin_location,
         category = category_name (executable_name))
      exception_origin_location_added = True
   else:
      information_message_box (
         title = application_title,
         message = unknown_exception_origin_message)
      exception_origin_location_added = False
   
# ============================================================================

def invok_addr2line (executable_name = None, exception_log = None):
   """
      + invok_addr2line (executable_name = None, exception_log = None):
      - Invoks <invok_addr2line_with_parameters> as many times as
        many addresses was found in <exception_log>, using
        <executable_name>.
      - If either <executable_name> or <exception_log> is None, then
        will invok <request_addr2line_parameters_input> to complete
        requested parameter, using <executable_name> and <exception_log>
        as default parameters, whenever one is not None, which
        <request_addr2line_parameters_input> will use to initialize its
        input fields.
      - If all invocations of <invok_addr2line_with_parameters> returned
        False, then opens <warning_message_box> to tell the user something
        is presumably wrong.
   """

   global application_title
   global all_addr2line_invocations_failed_message
   global exception_origin_location_added
   global no_location_informations_at_all_message

   # -------------------------------------------------------------------------
   # TOC
   # -------------------------------------------------------------------------
   # Get executable_name and addresses
   # Iteratively invok addr2line on addresses
   # -------------------------------------------------------------------------

   # -------------------------------------------------------------------------
   # Get executable_name and addresses
   
   if (executable_name == None) or (exception_log == None):
      # We have to open the dialog to get all required parameters.
      parameters = request_addr2line_parameters_input (
         initial_executable_name = executable_name,
         initial_exception_log   = exception_log)

      if parameters != None:
         if type (parameters [1]).__name__ == "list":
            executable_name = parameters [0]
            addresses       = parameters [1]
            exception_log   = parameters [2]
         elif parameters [1] == an_exception_log_without_trace_back:
            # No trace, but still give exception origin location
            # in the Locations pane and return.
            executable_name = parameters [0]
            exception_log = parameters [2]
            add_exception_origin_location (
               executable_name = executable_name,
               exception_log = exception_log)
            if not exception_origin_location_added:
               information_message_box (
                  title = application_title,
                  message = no_location_informations_at_all_message)
            return
         else:
            # Should never occurs, handled here for completness
            # and safety.
            raise Exception ("internal_error")
      else:
         # The user canceled the dialog.
         return None
   else:
      # All parameters provided
      addresses = get_addresses_or_kind_from_exception_log (exception_log)
      if addresses == None:
         # Finally, addresse are missing, so still need to
         # open a dialog to request parameters.
         parameters = request_addr2line_parameters_input (
            initial_executable_name = executable_name,
            initial_exception_log   = exception_log)

         if parameters != None:
            executable_name = parameters [0]
            addresses       = parameters [1]
            exception_log   = parameters [2]
         else:
            # The user canceled the dialog.
            return None

   # -------------------------------------------------------------------------
   # Add exception origin location before trace back locations

   add_exception_origin_location (
      executable_name = executable_name,
      exception_log   = exception_log)
   
   # -------------------------------------------------------------------------
   # Iteratively invok addr2line on addresses

   status = False

   address_index = 1

   for address in addresses:

      if invok_addr2line_with_parameters (
         executable_name = executable_name,
         address         = address,
         index           = address_index):
         
         status = True
         
         address_index = address_index + 1
         # This is important to increase the index only if
         # invok_addr2line_with_parameters was successful, because
         # when it was not, this means no entry were added in the
         # Locations pane, and meaningful index, must be index
         # of items in the Location pane. Order will still be
         # ok, but missing indexes in the items list, will
         # look like something disturbing for the user.

   if status == False:
      # All invocations of addr2line failed, there must be something
      # wrong with either the executable which returned exception
      # trace (missing line numbers in executable ?) or with
      # addr2line (not in the ${PATH} ?).
      warning_message_box (
         title   = application_title,
         message = all_addr2line_invocations_failed_message)

   return None

# ============================================================================

def on_click_invok_addr2line (ui_control):
   """
      + on_click_invok_addr2line (ui_control):
      - Event hanler to be attached to a button or menu entry.
      - Will run <invok_addr2line>
   """

   # ui_control is not used

   invok_addr2line ()

# ============================================================================

def on_GPS_UI_ready (hook_name):
   """
      + on_GPS_UI_ready (hook_name):
      - Event handler to be attached to a GPS hook
      - This will create the menu entry for the main
        functionality of this module
   """

   menu_item = GPS.Menu.create (
      path        = "/Tools/Addr 2 Line 2 Locations",
      on_activate = on_click_invok_addr2line)

   # The menu item will always be enabled, so no need to have
   # a global variable.

# ============================================================================
# Integration into GPS

GPS.Hook ("gps_started").add (on_GPS_UI_ready)

# ============================================================================
# Output samples from Addr2Line
# (used as a reference for addr2line output interpretation)

# 00494DAA at tiny-texts.adb:0
# 00533F47 at l:/applications/gnat/include/asis/asis-iterator.adb:156
# 00536516 at c:/documents and settings/yannick/bureau/(...etc...)
# 7C817075 in ?? at ??:0
#
# Comments:
# - Notice in the first one, the null line number
# - Notice in the second one, which is a normal result, there is no col number
# - Notice in the third one, a normal result on Windows, space without quotes
# - Notice in the forth one, the question marks
# - For the output interpretation, will split into words, check if there is at
#   least three word, check if the length of the first word is at least 8, the
#   second word is either "at" or "in" and the line ends with a int. If none
#   of these is checked, then the line will not be interpreted. If it is,
#   checked, lines containing "?" or ending with an int which is zero will be
#   ignored.
# - For others we will get text start at the second word (the one after the
#   length of the first and second word added two space), and add ":1" as
#   column numbers for the GPS.Locations.parse function. Note:
#   GPS.Locations.parse requires to append a space and then after, a message
#   also, otherwise it does not recognize the string and adds nothing to
#   the Locations pane.

# ============================================================================
# Samples exception messages
# (used as a reference for Ada/GNAT exception messages interpretation)
#
# raised ASIS.EXCEPTIONS.ASIS_INAPPROPRIATE_ELEMENT : a4g-vcheck.adb:885
#
# Execution terminated by unhandled exception
# Exception name: ASIS.EXCEPTIONS.ASIS_INAPPROPRIATE_ELEMENT
# Message: a4g-vcheck.adb:885
# Call stack traceback locations:
# 0x494daa 0x4c620c 0x533f27 0x533f58 0x533de0 0x533de0 0x533de0 (...etc...)
#
# raised PROGRAM_ERROR : My custom exception message
#
# Execution terminated by unhandled exception
# Exception name: PROGRAM_ERROR
# Message: My custom exception message
# Call stack traceback locations:
# 0x531790 0x532161 0x530f04 0x538333 0x401235 0x401286 0x7c817075

# Comments:
# - The first one, is a standard message with an executable which was
#   not built with "Store call stack trace in exceptions".
# - The second one, is a standard message with an executable which, on the
#   opposite, was built with "Store call stack trace in exceptions".
# - The third one, is an exception raised with a custom message from an
#   executable not built with "Store call stack trace in exceptions".
# - The forth one, is an exception raised with a custom message from an
#   executable which was built with "Store call stack trace in exceptions".
# - Notice that custom message may possibly span multiple lines.
# - When call stack traceback is present, the relevant text starts with
#   "\nExecution terminated by unhandled exception".
# - When no call stack traceback is present, the relevant text starts with
#   "\nraised "
# - The list of address is a list of hexadecimal integer using the C
#   notation. The length is not always the same, but presumably, it should
#   contains at least two digits to be safe to accept as an address.
# - The message part of the exception message, may be after the
#   "Exception name:" part and the "Message:" part together, ot the
#   whole exception message of the form "raised <NAME> : blablabla"
# - Blablabla or "Message:", may be a file name.
# - Blablabla or "Message:", when not a file name, may probably be
#   possibly multiple lines.

# ============================================================================
# Change log [yannick_duchene@yahoo.fr or yannick-duchene@rasama.org]
#
# ----------------------------------------------------------------------------
# 2010-01-16 : First revision and initial release
# - Initial release, posted on comp.lang.ada and fr.comp.lang.ada
# - http://groups.google.com/group/comp.lang.ada/browse_thread/thread/475d5d08fbf558a2?hl=fr#
# - http://groups.google.com/group/fr.comp.lang.ada/browse_thread/thread/6ec414c9dd4a9f23?hl=fr#
#
# ----------------------------------------------------------------------------
# 2010-01-17 : Second revision and online documentations
# - Corrected a potential bug : the addr2line --exe parameter now
#   enclosed in quotes.
# - The dialog exception log text field is now automatically initialized after
#   the text contained in any console active at the moment the command is
#   invoked.
# - The dialog executable name is now automatically initialized after
#   project properties.
# - The methods are now aware of the message contained in exception log text
#   and may add an entry in Locations for the location of the instruction
#   which raised an exception.
# - Related to the latter point, the dialog may now accept an exception
#   log text event if it contains no traceback list, providing it
#   contains at least a short form exception log (the one you get when
#   an application is built without the exception traceback option enabled).
# - Numerous renaming so that entity in this script match the new
#   specifications.
# - Merged all text table sections into a single one at the top of the file.
